# Copyright 2022 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

"""`wandbfsspec `: fsspec interface for Weights & Biases (wandb)"""

__author__ = "Alvaro Bartolome, alvarobartt @ GitHub"
__version__ = "0.1.0"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['wandbfsspec']

package_data = \
{'': ['*']}

install_requires = \
['fsspec>=2022.5.0,<2023.0.0', 'wandb>=0.12.20,<0.13.0']

entry_points = \
{'fsspec.specs': ['wandbfs = wandbfsspec.core.WandbFileSystem']}

setup_kwargs = {
    'name': 'wandbfsspec',
    'version': '0.1.0',
    'description': 'fsspec interface for Weights & Biases (wandb)',
    'long_description': '# 🍱 `fsspec` interface for Weights & Biases (wandb)\n\nQuoting Weights and Biases (wandb), "Weights & Biases is the \nmachine learning platform for developers to build better models \nfaster. Use W&B\'s lightweight, interoperable tools to quickly track \nexperiments, version and iterate on datasets, evaluate model performance, \nreproduce models, visualize results and spot regressions, and share \nfindings with colleagues.". Reference at https://docs.wandb.ai/\n\nSo you may be thinking, what does `wandb` have to do with anything\nclose to a File System? Well, it\'s not but it actually provides a way\nto upload/download files and store them in a remote, which makes it somehow\na File System. Also, `wandb` provides an API that lets you interact with\nthat "File System", so this is why `wandbfsspec` makes sense, in order to ease\nthat interface between `wandb`\'s File System and anyone willing to use it.\n\nThe `wandbfsspec` implementation is based on https://github.com/fsspec/filesystem_spec.\n\n## 🔮 Future TODOs\n\nObviously, since `wandb`\'s main purpose is to track and monitor ML experiments,\nit contains an artifact store, so as to dump there the experiment artifacts for data\nversioning and model tracking. More information in https://wandb.ai/site/artifacts.\n\nSo on, a new interface will be implemented in `wandbfsspec` not just to handle the files\nthat can be uploaded/downloaded to/from `wandb`, but also the artifacts. So the next release\nwill implement a new `AbstractFileSystem` class named `WandbArtifactStore` with the protocol\n`wandbas` in order to access the artifact store as if it was a default File System.\n\nSome more notes on how to actually use `wandb`\'s artifact store at https://docs.wandb.ai/guides/artifacts.\n\nOnce that\'s done, we\'ll fill a PR in https://github.com/fsspec/filesystem_spec, so as to\nregister both protocols supported by `wandbfsspec`: `wandbfs` and `wandbas`.\n\n## 🚸 Usage\n\nHere\'s an example on how to locate and open a file:\n\n```python\n>>> from wandbfsspec.core import WandbFileSystem\n>>> fs = WandbFileSystem(api_key="YOUR_API_KEY")\n>>> fs.ls("alvarobartt/wandbfsspec-tests/3s6km7mp")\n[\'alvarobartt/wandbfsspec-tests/3s6km7mp/config.yaml\', \'alvarobartt/wandbfsspec-tests/3s6km7mp/file.yaml\', \'alvarobartt/wandbfsspec-tests/3s6km7mp/files\', \'alvarobartt/wandbfsspec-tests/3s6km7mp/output.log\', \'alvarobartt/wandbfsspec-tests/3s6km7mp/requirements.txt\', \'alvarobartt/wandbfsspec-tests/3s6km7mp/wandb-metadata.json\', \'alvarobartt/wandbfsspec-tests/3s6km7mp/wandb-summary.json\']\n>>> with fs.open("alvarobartt/wandbfsspec-tests/3s6km7mp/file.yaml", "rb") as f:\n...     print(f.read())\nb\'some: data\\nfor: testing\'\n```\n\n📌 Note that it can also be done through `fsspec` as long as `wandbfsspec` is installed:\n\n```python\n>>> import fsspec\n>>> fs = fsspec.filesystem("wandbfs")\n>>> fs.ls("alvarobartt/wandbfsspec-tests/3s6km7mp")\n[\'alvarobartt/wandbfsspec-tests/3s6km7mp/config.yaml\', \'alvarobartt/wandbfsspec-tests/3s6km7mp/file.yaml\', \'alvarobartt/wandbfsspec-tests/3s6km7mp/files\', \'alvarobartt/wandbfsspec-tests/3s6km7mp/output.log\', \'alvarobartt/wandbfsspec-tests/3s6km7mp/requirements.txt\', \'alvarobartt/wandbfsspec-tests/3s6km7mp/wandb-metadata.json\', \'alvarobartt/wandbfsspec-tests/3s6km7mp/wandb-summary.json\']\n>>> with fs.open("alvarobartt/wandbfsspec-tests/3s6km7mp/file.yaml", "rb") as f:\n...     print(f.read())\nb\'some: data\\nfor: testing\'\n```\n\n## 📝 Documentation\n\nComing soon... (https://github.com/mkdocs/mkdocs)\n\n## 🧪 How to test it\n\nIn order to test it, you should first set the following environment variables\nso as to use `wandb` as a file system for the tests.\n\n```\nWANDB_ENTITY = ""\nWANDB_PROJECT = ""\nWANDB_API_KEY = ""\n```\n\nBoth entity and project values can be found in your https://wandb.ai/ account, as\nthe entity name is your account name, and the project name can either be already\ncreated or you can just specify it and it\'ll be created during `pytest` init. Then,\nregarding the API Key, you just need to go to https://wandb.ai/settings, scroll\ndown to Danger Zone -> API Keys, and copy your personal API Key from there.\n\n⚠️ Make sure that you don\'t publish your API Key anywhere, that\'s why we\'re defining\nit as an environment value, so as to avoid potential issues on commiting code with\nthe actual API Key value.\n\nThen, in order to actually run the tests you can either run:\n\n- `poetry run pytest`\n- `poetry run make tests`\n\nOr, if you\'re not using `poetry`, you can just run both those commands without it.',
    'author': 'Alvaro Bartolome',
    'author_email': 'alvarobartt@yahoo.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alvarobartt/wandbfsspec',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



# Copyright 2022 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

import datetime
import os
import tempfile
import urllib.request
import warnings
from pathlib import Path
from typing import Any, Dict, List, Literal, Tuple, Union

import wandb
from fsspec import AbstractFileSystem
from fsspec.spec import AbstractBufferedFile

MAX_PATH_LENGTH_WITHOUT_FILE_PATH = 3


class WandbFileSystem(AbstractFileSystem):
    protocol = "wandbfs"

    def __init__(
        self,
        api_key: Union[str, None] = None,
    ) -> None:
        super().__init__()

        if api_key:
            os.environ["WANDB_API_KEY"] = api_key

        assert os.getenv("WANDB_API_KEY"), (
            "In order to connect to the wandb Public API you need to provide the API"
            " key either via param `api_key`, setting the key in the environment"
            " variable `WANDB_API_KEY`, or running `wandb login <WANDB_API_KEY>`."
        )

        self.api = wandb.Api()

    @classmethod
    def split_path(
        self, path: str
    ) -> Tuple[str, Union[str, None], Union[str, None], Union[str, None]]:
        path = self._strip_protocol(path=path)
        path = path.lstrip("/")
        if "/" not in path:
            return (path, None, None, None)
        path = path.split("/")
        if len(path) > MAX_PATH_LENGTH_WITHOUT_FILE_PATH:
            return (
                *path[:MAX_PATH_LENGTH_WITHOUT_FILE_PATH],
                "/".join(path[MAX_PATH_LENGTH_WITHOUT_FILE_PATH:]),
            )
        path += [None] * (MAX_PATH_LENGTH_WITHOUT_FILE_PATH - len(path))
        return (*path, None)

    def ls(self, path: str, detail: bool = False) -> Union[List[str], Dict[str, Any]]:
        entity, project, run_id, file_path = self.split_path(path=path)
        if entity and project and run_id:
            _files = self.api.run(f"{entity}/{project}/{run_id}").files()
            base_path = f"{entity}/{project}/{run_id}"
            return self.__ls_files(
                _files=_files,
                base_path=f"{base_path}/{file_path}" if file_path else base_path,
                file_path=file_path if file_path else Path("./"),
                detail=detail,
            )
        elif entity and project:
            _files = self.api.runs(f"{entity}/{project}")
            base_path = f"{entity}/{project}"
            return self.__ls_projects_or_runs(
                _files=_files, base_path=base_path, detail=detail
            )
        elif entity:
            _files = self.api.projects(entity=entity)
            base_path = entity
            return self.__ls_projects_or_runs(
                _files=_files, base_path=base_path, detail=detail
            )
        return []

    @staticmethod
    def __ls_files(
        _files: List[str],
        base_path: Union[str, Path],
        file_path: Union[str, Path] = Path("./"),
        detail: bool = False,
    ) -> Union[List[str], Dict[str, Any]]:
        file_path = Path(file_path) if isinstance(file_path, str) else file_path
        files = []
        for _file in _files:
            if file_path not in Path(_file.name).parents:
                continue
            filename = Path(_file.name.replace(f"{file_path.as_posix()}/", ""))
            if filename.is_dir() or len(filename.parents) > 1:
                filename = filename.parent
                if (
                    any(f["name"] == f"{base_path}/{filename.name}" for f in files)
                    if detail
                    else f"{base_path}/{filename.name}" in files
                ):
                    continue
                files.append(
                    {
                        "name": f"{base_path}/{filename.name}",
                        "type": "directory",
                        "size": 0,
                    }
                    if detail
                    else f"{base_path}/{filename.name}"
                )
                continue
            files.append(
                {
                    "name": f"{base_path}/{filename.name}",
                    "type": "file",
                    "size": _file.size,
                }
                if detail
                else f"{base_path}/{filename.name}"
            )
        return files

    @staticmethod
    def __ls_projects_or_runs(
        _files: List[str], detail: bool = False
    ) -> Union[List[str], Dict[str, Any]]:
        files = []
        for _file in _files:
            files.append(
                {
                    "name": _file.name,
                    "type": "directory",
                    "size": 0,
                }
                if detail
                else _file.name
            )
        return files

    def modified(self, path: str) -> datetime.datetime:
        """Return the modified timestamp of a file as a datetime.datetime"""
        entity, project, run_id, file_path = self.split_path(path=path)
        if not file_path:
            raise ValueError
        _file = self.api.run(f"{entity}/{project}/{run_id}").file(name=file_path)
        if not _file:
            raise ValueError
        return datetime.datetime.fromisoformat(_file.updated_at)

    def open(self, path: str, mode: Literal["rb", "wb"] = "rb") -> None:
        _, _, _, file_path = self.split_path(path=path)
        if not file_path:
            raise ValueError
        return WandbFile(self, path=path, mode=mode)

    def url(self, path: str) -> str:
        entity, project, run_id, file_path = self.split_path(path=path)
        _file = self.api.run(f"{entity}/{project}/{run_id}").file(name=file_path)
        if not _file:
            raise ValueError
        return _file.direct_url

    def cat_file(
        self, path: str, start: Union[int, None] = None, end: Union[int, None] = None
    ) -> bytes:
        url = self.url(path=path)
        req = urllib.request.Request(url=url)
        if not start and not end:
            start, end = 0, ""
        req.add_header("Range", f"bytes={start}-{end}")
        return urllib.request.urlopen(req).read()

    def put_file(self, lpath: str, rpath: str, **kwargs) -> None:
        entity, project, run_id, file_path = self.split_path(path=rpath)
        run = self.api.run(f"{entity}/{project}/{run_id}")
        warnings.warn(
            "`rpath` should be a directory path not a file path, as in order to use"
            " file paths we'll need to wait upon"
            " https://github.com/wandb/client/pull/3929 merge",
            RuntimeWarning,
        )
        run.upload_file(path=lpath, root=file_path if file_path else ".")

    def get_file(
        self, lpath: str, rpath: str, overwrite: bool = False, **kwargs
    ) -> None:
        entity, project, run_id, file_path = self.split_path(path=rpath)
        file = self.api.run(f"{entity}/{project}/{run_id}").file(name=file_path)
        file.download(root=lpath, replace=overwrite)

    def rm_file(self, path: str) -> None:
        entity, project, run_id, file_path = self.split_path(path=path)
        file = self.api.run(f"{entity}/{project}/{run_id}").file(name=file_path)
        file.delete()

    def cp_file(self, path1: str, path2: str, **kwargs) -> None:
        warnings.warn(
            "`path2` should be a directory path not a file path, as in order to use"
            " file paths we'll need to wait upon"
            " https://github.com/wandb/client/pull/3924 merge",
            RuntimeWarning,
        )
        # with tempfile.NamedTemporaryFile() as f: f.name
        with tempfile.TemporaryDirectory() as f:
            self.get_file(lpath=f, rpath=path1, overwrite=True)
            _, _, _, file_path = self.split_path(path=path1)
            self.put_file(lpath=f"{f}/{file_path}", rpath=path2)


class WandbFile(AbstractBufferedFile):
    def __init__(
        self, fs: WandbFileSystem, path: str, mode: Literal["rb", "wb"] = "rb"
    ) -> None:
        super().__init__(fs=fs, path=path, mode=mode)

    def _fetch_range(
        self, start: Union[int, None] = None, end: Union[int, None] = None
    ) -> bytes:
        return self.fs.cat_file(path=self.path, start=start, end=end)
